from django.test import AsyncClient, Client


def test_say_hello(client: Client) -> None:
    response = client.get("/api/say-hello", data={"name": "World"})

    assert response.status_code == 200
    assert response.json() == "Hello, World!"


async def test_async_say_hello(async_client: AsyncClient) -> None:
    response = await async_client.get("/api/say-hello-async", data={"name": "World"})

    assert response.status_code == 200
    assert response.json() == "Hello, World!"



from django.test import AsyncClient, Client


def test_injected_setting(client: Client) -> None:
    response = client.get("/get-setting/")

    assert response.status_code == 200
    assert response.content == b"Hello, World!"


async def test_injected_setting_async(async_client: AsyncClient) -> None:
    response = await async_client.get("/get-setting-async/")

    assert response.status_code == 200
    assert response.content == b"Hello, World!"


def test_configure_dependency(client: Client) -> None:
    response = client.get("/get-configured-dependency/")

    assert response.status_code == 200
    assert response.content == b"This is configured string"


def test_request_scoped_dependency(client: Client) -> None:
    response = client.get("/get-request-scoped-dependency/")

    assert response.status_code == 200

    request_id = response.content

    response = client.get("/get-request-scoped-dependency/")

    next_request_id = response.content

    assert request_id != next_request_id


async def test_request_scoped_dependency_async(async_client: AsyncClient) -> None:
    response = await async_client.get("/get-request-scoped-dependency-async/")

    assert response.status_code == 200

    request_id = response.content

    response = await async_client.get("/get-request-scoped-dependency-async/")

    next_request_id = response.content

    assert request_id != next_request_id



from typing import Annotated

from django.http import HttpRequest, HttpResponse

import anydi


def get_setting(
    request: HttpRequest,
    message: Annotated[str, "django.conf.settings.HELLO_MESSAGE"] = anydi.auto,
) -> HttpResponse:
    return HttpResponse(message)


async def get_setting_async(
    request: HttpRequest,
    message: Annotated[str, "django.conf.settings.HELLO_MESSAGE"] = anydi.auto,
) -> HttpResponse:
    return HttpResponse(message)


def get_configured_dependency(
    request: HttpRequest,
    message: Annotated[str, "configured-string"] = anydi.auto,
) -> HttpResponse:
    return HttpResponse(message)


def get_request_scoped_dependency(
    request: HttpRequest,
    request_id: Annotated[str, "request-id"] = anydi.auto,
) -> HttpResponse:
    return HttpResponse(request_id)


async def get_request_scoped_dependency_async(
    request: HttpRequest,
    request_id: Annotated[str, "request-id"] = anydi.auto,
) -> HttpResponse:
    return HttpResponse(request_id)



from typing import Any

from django.http import HttpRequest
from ninja import Router

import anydi

from tests.ext.django.services import HelloService

router = Router()


@router.get("/say-hello")
def say_hello(
    request: HttpRequest,
    name: str,
    hello_service: HelloService = anydi.auto,
) -> Any:
    assert hello_service.started
    return hello_service.say_hello(name)


@router.get("/say-hello-async")
async def say_hello_async(
    request: HttpRequest,
    name: str,
    hello_service: HelloService = anydi.auto,
) -> Any:
    assert hello_service.started
    return await hello_service.say_hello_async(name)



import pytest

from anydi import Container
from anydi.ext import pytest_plugin


class Service:
    pass


class UnknownService:
    pass


@pytest.fixture(scope="module")
def container() -> Container:
    container = Container()
    container.register(Service, lambda: Service(), scope="singleton")
    return container


def test_anydi_inject_all_default(request: pytest.FixtureRequest) -> None:
    assert request.config.getini("anydi_inject_all") is False


def test_no_container_setup(
    request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(pytest_plugin, "CONTAINER_FIXTURE_NAME", "container1")

    with pytest.raises(pytest.FixtureLookupError) as exc_info:
        request.getfixturevalue("anydi_setup_container")

    assert exc_info.value.msg == (
        "`container` fixture is not found. Make sure to define it in your test module "
        "or override `anydi_setup_container` fixture."
    )


@pytest.mark.inject
def test_inject_service(service: Service) -> None:
    assert isinstance(service, Service)


@pytest.mark.xfail
@pytest.mark.inject
def test_inject_unknown_service(unknown_service: UnknownService) -> None:
    pass


@pytest.mark.inject
async def test_ainject_service(service: Service) -> None:
    assert isinstance(service, Service)


@pytest.mark.xfail
@pytest.mark.inject
async def test_ainject_unknown_service(unknown_service: UnknownService) -> None:
    pass


@pytest.mark.xfail
@pytest.mark.inject
def test_inject_missing_type(service) -> None:  # type: ignore[no-untyped-def]
    pass



import anydi

from tests.fixtures import Service


class ScanAppModule(anydi.Module):
    @anydi.provider(scope="singleton")
    def a_a1_provider(self) -> str:
        return "a.a1.str_provider"

    @anydi.provider(scope="singleton")
    def a_a3_provider(self) -> int:
        return 10000

    @anydi.provider(scope="singleton")
    def b_service_provider(self, ident: str) -> Service:
        return Service(ident=ident)



import uuid
from collections.abc import AsyncIterator, Iterator
from dataclasses import dataclass, field
from typing import Annotated

from anydi import Module, provider
from anydi._container import Container


def func() -> str:
    return "func"


class Class:
    pass


def generator() -> Iterator[str]:
    yield "generator"


async def async_generator() -> AsyncIterator[str]:
    yield "async_generator"


async def coro() -> str:
    return "coro"


def event() -> Iterator[None]:
    yield


async def async_event() -> AsyncIterator[None]:
    yield


def iterator() -> Iterator:  # type: ignore[type-arg]
    yield


@dataclass(frozen=True)
class UniqueId:
    id: uuid.UUID = field(default_factory=uuid.uuid4)


class Service:
    def __init__(self, ident: str) -> None:
        self.ident = ident
        self.events: list[str] = []


class Resource:
    def __init__(self) -> None:
        self.called = False
        self.committed = False
        self.rolled_back = False

    def run(self) -> None:
        self.called = True

    def commit(self) -> None:
        self.committed = True

    def rollback(self) -> None:
        self.rolled_back = True


class TestModule(Module):
    def configure(self, container: Container) -> None:
        container.register(
            Annotated[str, "msg1"], lambda: "Message 1", scope="singleton"
        )

    @provider(scope="singleton")
    def provide_msg2(self) -> Annotated[str, "msg2"]:
        return "Message 2"



from typing import Annotated, Any

from fastapi import Body, Depends, FastAPI
from starlette.middleware import Middleware

import anydi.ext.fastapi
from anydi.ext.fastapi import Inject
from anydi.ext.starlette.middleware import RequestScopedMiddleware
from anydi.testing import TestContainer

from tests.ext.fixtures import Mail, MailService, User, UserService

container = TestContainer()


@container.provider(scope="singleton")
def message1() -> Annotated[str, "message1"]:
    return "message1"


@container.provider(scope="singleton")
def message1_a() -> Annotated[str, "message1", "a"]:
    return "message1_a"


@container.provider(scope="singleton")
def message1_a_b() -> Annotated[str, "message1", "a", "b"]:
    return "message1_a_b"


@container.provider(scope="singleton")
def message2() -> Annotated[str, "message2"]:
    return "message2"


@container.provider(scope="singleton")
def user_service() -> UserService:
    return UserService()


@container.provider(scope="request")
def mail_service() -> MailService:
    return MailService()


async def get_user(user_service: UserService = Inject()) -> User:
    return await user_service.get_user()


app = FastAPI(middleware=[Middleware(RequestScopedMiddleware, container=container)])


@app.post("/send-mail", response_model=Mail)
async def send_email(
    user: User = Depends(get_user),
    mail_service: MailService = Inject(),
    message: str = Body(embed=True),
) -> Any:
    return await mail_service.send_mail(email=user.email, message=message)


@app.post("/send-mail-annotated", response_model=Mail)
async def send_email_annotated(
    user: Annotated[User, Depends(get_user)],
    mail_service: Annotated[MailService, Inject()],
    message: Annotated[str, Body(embed=True)],
) -> Any:
    return await mail_service.send_mail(email=user.email, message=message)


@app.get("/annotated-mixed")
def annotated_mixed(
    message1: Annotated[Annotated[str, "message1"], Inject()],
    message1_a: Annotated[Annotated[str, "message1", "a"], Inject()],
    message1_a_b: Annotated[Annotated[str, "message1", "a", "b"], Inject()],
    message2: Annotated[str, "message2"] = Inject(),
) -> Any:
    return [
        message1,
        message1_a,
        message1_a_b,
        message2,
    ]


anydi.ext.fastapi.install(app, container)



SECRET_KEY = "secret"
DEBUG = True

INSTALLED_APPS = ("anydi.ext.django",)

ROOT_URLCONF = "tests.ext.django.urls"

MIDDLEWARE = ("anydi.ext.django.middleware.request_scoped_middleware",)

# AnyDI settings
ANYDI = {
    "STRICT_MODE": False,
    "REGISTER_SETTINGS": True,
    "REGISTER_COMPONENTS": True,
    "MODULES": ["tests.ext.django.container.configure"],
    "PATCH_NINJA": True,
    "INJECT_URLCONF": ROOT_URLCONF,
    "SCAN_PACKAGES": ["tests.ext.django.scan"],
}

# Custom settings
HELLO_MESSAGE = "Hello, World!"



from anydi import Module, injectable, provider, request, singleton, transient
from anydi._decorators import is_injectable, is_provided

from tests.fixtures import Service


def test_provider_decorator() -> None:
    class TestModule(Module):
        @provider(scope="singleton", override=True)
        def provider(self) -> str:
            return "test"

    assert getattr(TestModule.provider, "__provider__") == {
        "scope": "singleton",
        "override": True,
    }


def test_request_decorator() -> None:
    request(Service)

    assert is_provided(Service)
    assert Service.__provided__ == {
        "scope": "request",
    }


def test_transient_decorator() -> None:
    transient(Service)

    assert is_provided(Service)
    assert Service.__provided__ == {
        "scope": "transient",
    }


def test_singleton_decorator() -> None:
    singleton(Service)

    assert is_provided(Service)
    assert Service.__provided__ == {
        "scope": "singleton",
    }


def test_injectable_no_args() -> None:
    @injectable
    def my_func() -> None:
        pass

    assert is_injectable(my_func)
    assert my_func.__injectable__ == {
        "tags": None,
    }


def test_injectable_no_args_provided() -> None:
    @injectable()
    def my_func() -> None:
        pass

    assert is_injectable(my_func)
    assert my_func.__injectable__ == {
        "tags": None,
    }


def test_injectable_with_tags() -> None:
    @injectable(tags=["tag1", "tag2"])
    def my_func() -> None:
        pass

    assert is_injectable(my_func)
    assert my_func.__injectable__ == {
        "tags": ["tag1", "tag2"],
    }



from __future__ import annotations

import enum
import inspect
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Callable, NamedTuple

from ._scope import Scope
from ._typing import NOT_SET


class ProviderKind(enum.IntEnum):
    CLASS = 1
    FUNCTION = 2
    COROUTINE = 3
    GENERATOR = 4
    ASYNC_GENERATOR = 5

    @classmethod
    def from_call(cls, call: Callable[..., Any]) -> ProviderKind:
        if inspect.isclass(call):
            return cls.CLASS
        if inspect.iscoroutinefunction(call):
            return cls.COROUTINE
        if inspect.isasyncgenfunction(call):
            return cls.ASYNC_GENERATOR
        if inspect.isgeneratorfunction(call):
            return cls.GENERATOR
        if inspect.isfunction(call) or inspect.ismethod(call):
            return cls.FUNCTION
        raise TypeError(
            f"The provider `{call}` is invalid because it is not a callable object."
        )

    @classmethod
    def is_resource(cls, kind: ProviderKind) -> bool:
        return kind in (cls.GENERATOR, cls.ASYNC_GENERATOR)


@dataclass(kw_only=True, frozen=True)
class Provider:
    call: Callable[..., Any]
    scope: Scope
    interface: Any
    name: str
    parameters: list[inspect.Parameter]
    kind: ProviderKind

    def __str__(self) -> str:
        return self.name

    @cached_property
    def is_class(self) -> bool:
        return self.kind == ProviderKind.CLASS

    @cached_property
    def is_coroutine(self) -> bool:
        return self.kind == ProviderKind.COROUTINE

    @cached_property
    def is_generator(self) -> bool:
        return self.kind == ProviderKind.GENERATOR

    @cached_property
    def is_async_generator(self) -> bool:
        return self.kind == ProviderKind.ASYNC_GENERATOR

    @cached_property
    def is_async(self) -> bool:
        return self.is_coroutine or self.is_async_generator

    @cached_property
    def is_resource(self) -> bool:
        return ProviderKind.is_resource(self.kind)


class ProviderDef(NamedTuple):
    call: Callable[..., Any]
    scope: Scope
    interface: Any = NOT_SET



from typing import Annotated

import pytest

from anydi import Container, Module, provider

from tests.fixtures import TestModule


class TestContainerModuleRegistrator:
    @pytest.fixture
    def container(self) -> Container:
        return Container()

    def test_register_modules(self) -> None:
        container = Container(modules=[TestModule])

        assert container.is_registered(Annotated[str, "msg1"])
        assert container.is_registered(Annotated[str, "msg2"])

    def test_register_module_class(self, container: Container) -> None:
        container.register_module(TestModule)

        assert container.is_registered(Annotated[str, "msg1"])
        assert container.is_registered(Annotated[str, "msg2"])

    def test_register_module_instance(self, container: Container) -> None:
        container.register_module(TestModule())

        assert container.is_registered(Annotated[str, "msg1"])
        assert container.is_registered(Annotated[str, "msg2"])

    def test_register_module_path(self, container: Container) -> None:
        container.register_module("tests.test_module.TestModule")

        assert container.is_registered(Annotated[str, "msg1"])
        assert container.is_registered(Annotated[str, "msg2"])

    def test_register_module_function(self, container: Container) -> None:
        def configure(_container: Container) -> None:
            _container.register(str, lambda: "Message 1", scope="singleton")

        container.register_module(configure)

        assert container.is_registered(str)

    def test_register_module_ordered_providers(self, container: Container) -> None:
        class OrderedModule(Module):
            @provider(scope="singleton")
            def dep3(self) -> Annotated[str, "dep3"]:
                return "dep3"

            @provider(scope="singleton")
            def dep1(self) -> Annotated[str, "dep1"]:
                return "dep1"

            @provider(scope="singleton")
            def dep2(self) -> Annotated[str, "dep2"]:
                return "dep2"

        container.register_module(OrderedModule)

        assert list(container.providers.keys()) == [
            Annotated[str, "dep3"],
            Annotated[str, "dep1"],
            Annotated[str, "dep2"],
        ]

    def test_register_module_invalid_path(self, container: Container) -> None:
        with pytest.raises(
            TypeError,
            match="The module must be a callable, a module type, or a module instance.",
        ):
            container.register_module("anydi.Container")



from dataclasses import dataclass

TEST_EMAIL = "test@mail.com"


@dataclass
class User:
    id: int
    email: str


class UserService:
    def __init__(self) -> None:
        self._users: dict[int, User] = {}

    async def create_user(self, id_: int, email: str) -> User:
        user = User(id=id_, email=email)
        self._users[id_] = user
        return user

    async def get_user_by_id(self, id_: int) -> User:
        return self._users[id_]

    async def get_user(self) -> User:
        return User(id=1, email=TEST_EMAIL)


@dataclass
class Mail:
    email: str
    message: str


class MailService:
    async def send_mail(self, email: str, message: str) -> Mail:
        return Mail(email=email, message=message)



import sys
from typing import Annotated, Any, Union

import pytest

from anydi._typing import is_builtin_type, type_repr

from tests.fixtures import Service


@pytest.mark.parametrize(
    ("tp", "expected"),
    [
        (bool, True),
        (str, True),
        (int, True),
        (float, True),
        (Service, False),
    ],
)
def test_is_builtin_type(tp: type[Any], expected: bool) -> None:
    assert is_builtin_type(tp) == expected


@pytest.mark.parametrize(
    ("obj", "expected_qualname"),
    [
        (int, "int"),
        (Service, "tests.fixtures.Service"),
        (Service(ident="test"), "tests.fixtures.Service"),
        pytest.param(
            Annotated[Service, "service"],
            'Annotated[tests.fixtures.Service, "service"]',
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 9), reason="Requires Python 3.9"
            ),
        ),
        (lambda x: x, "tests.test_utils.<lambda>"),
        (123, "int"),
        ("hello", "hello"),
        (list[str], "list[str]"),
        pytest.param(
            Union[str, int],
            "Union[str, int]",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 10), reason="Requires Python 3.10"
            ),
        ),
        pytest.param(
            Union[str, int],
            "_SpecialForm[str, int]",
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 10), reason="Requires Python 3.9"
            ),
        ),
    ],
)
def test_type_repr(obj: Any, expected_qualname: str) -> None:
    qualname = type_repr(obj)

    assert qualname == expected_qualname



from typing import Annotated

from pydantic import computed_field
from pydantic_settings import BaseSettings

import anydi.ext.pydantic_settings
from anydi import Container


class Settings(BaseSettings):
    param_str: str = "test"
    param_int: int = 42
    param_float: float = 3.14

    @computed_field
    def param_computed(self) -> str:
        return "computed"


class DBSettings(BaseSettings):
    db_url: str = "sqlite://:memory:"


def test_install_settings() -> None:
    container = Container()

    anydi.ext.pydantic_settings.install(Settings(), container, prefix="settings")

    assert container.resolve(Annotated[str, "settings.param_str"]) == "test"
    assert container.resolve(Annotated[int, "settings.param_int"]) == 42
    assert container.resolve(Annotated[float, "settings.param_float"]) == 3.14
    assert container.resolve(Annotated[str, "settings.param_computed"]) == "computed"


def test_install_multiple_settings() -> None:
    container = Container()

    anydi.ext.pydantic_settings.install(
        [Settings(), DBSettings()], container, prefix="settings."
    )

    assert container.resolve(Annotated[str, "settings.param_str"]) == "test"
    assert container.resolve(Annotated[str, "settings.db_url"]) == "sqlite://:memory:"



from __future__ import annotations

from types import TracebackType

from typing_extensions import Self


class HelloService:
    def __init__(self) -> None:
        self.started = False

    def say_hello(self, name: str) -> str:
        return f"Hello, {name}!"

    async def say_hello_async(self, name: str) -> str:
        return f"Hello, {name}!"

    def __enter__(self) -> Self:
        self.started = True
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.started = False



from typing import Any

import pytest
from faststream.redis import RedisBroker, TestRedisBroker
from pydantic import BaseModel

import anydi.ext.faststream
from anydi import Container
from anydi.ext.faststream import Inject

from tests.ext.fixtures import UserService


class UserMessage(BaseModel):
    id: int
    email: str


@pytest.fixture(scope="session")
def container() -> Container:
    container = Container()

    @container.provider(scope="singleton")
    def provide_user_service() -> UserService:
        return UserService()

    return container


@pytest.fixture(scope="session")
def broker(container: Container) -> RedisBroker:
    broker = RedisBroker()

    @broker.subscriber("test-in")
    async def create_user(
        message: UserMessage,
        user_service: UserService = Inject(),
    ) -> Any:
        return await user_service.create_user(id_=message.id, email=message.email)

    anydi.ext.faststream.install(broker, container)

    return broker


async def test_handle(broker: RedisBroker, container: Container) -> None:
    user_service = container.resolve(UserService)

    async with TestRedisBroker(broker) as br:
        await br.publish(UserMessage(id=1, email="test@mail.com"), channel="test-in")

    user = await user_service.get_user_by_id(1)

    assert user.email == "test@mail.com"



import pytest
from pytest_mock import MockerFixture

from anydi import Container

from tests.scan_app import ScanAppModule


class TestContainerScan:
    @pytest.fixture
    def container(self) -> Container:
        return Container()

    def test_scan_registers_all_dependencies(
        self, container: Container, mocker: MockerFixture
    ) -> None:
        inject_spy = mocker.spy(container, "inject")

        container.register_module(ScanAppModule)
        container.scan(["tests.scan_app"])

        # Expecting 7 total inject calls from all eligible @injectable functions
        assert inject_spy.call_count == 7

        from .scan_app.a.a3.handlers import a_a3_handler_1, a_a3_handler_2

        assert a_a3_handler_1() == "a.a1.str_provider"
        assert a_a3_handler_2().ident == "a.a1.str_provider"

    def test_scan_single_module_registers_limited_dependencies(
        self, container: Container, mocker: MockerFixture
    ) -> None:
        inject_spy = mocker.spy(container, "inject")

        container.register_module(ScanAppModule)
        container.scan("tests.scan_app.a.a3.handlers")

        # Expecting only 2 handlers to be registered from this specific module
        assert inject_spy.call_count == 2

        from .scan_app.a.a3.handlers import a_a3_handler_1

        assert a_a3_handler_1() == "a.a1.str_provider"

    def test_scan_with_unknown_tag_registers_nothing(
        self, container: Container, mocker: MockerFixture
    ) -> None:
        inject_spy = mocker.spy(container, "inject")

        container.scan(["tests.scan_app"], tags=["non_existing_tag"])

        # No injectable functions should be picked up with unmatched tags
        inject_spy.assert_not_called()

    def test_scan_with_matching_tag_registers_only_matching_dependencies(
        self, container: Container, mocker: MockerFixture
    ) -> None:
        inject_spy = mocker.spy(container, "inject")

        container.register_module(ScanAppModule)
        container.scan(["tests.scan_app"], tags=["inject"])

        # Only handlers with @injectable(..., tags=["inject"]) should be included
        assert inject_spy.call_count == 1

        from .scan_app.a.a3.handlers import a_a3_handler_1

        assert a_a3_handler_1() == "a.a1.str_provider"



from typing import cast

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.routing import Route

from anydi import auto
from anydi.ext.starlette.middleware import RequestScopedMiddleware
from anydi.testing import TestContainer

from tests.ext.fixtures import MailService, UserService

container = TestContainer()


@container.provider(scope="singleton")
def user_service() -> UserService:
    return UserService()


@container.provider(scope="singleton")
def mail_service() -> MailService:
    return MailService()


@container.inject
async def send_email(
    request: Request,
    user_service: UserService = auto,
    mail_service: MailService = auto,
) -> JSONResponse:
    data = await request.json()
    message = cast(str, data.get("message"))
    user = await user_service.get_user()
    mail = await mail_service.send_mail(email=user.email, message=message)
    return JSONResponse(
        {
            "email": user.email,
            "message": mail.message,
        }
    )


app = Starlette(
    routes=[
        Route("/send-mail", send_email, methods=["POST"]),
    ],
    middleware=[Middleware(RequestScopedMiddleware, container=container)],
)

app.state.container = container



from unittest import mock

from starlette.testclient import TestClient

from anydi.testing import TestContainer

from tests.ext.fixtures import TEST_EMAIL, Mail, MailService, User, UserService


def test_send_mail(client: TestClient) -> None:
    message = "test"

    response = client.post("/send-mail", json={"message": message})

    assert response.status_code == 200
    assert response.json() == {
        "email": TEST_EMAIL,
        "message": message,
    }


def test_send_mail_mock_mail_service(
    client: TestClient, container: TestContainer
) -> None:
    mail = Mail(email=TEST_EMAIL, message="mock")

    mail_service_mock = mock.MagicMock(spec=MailService)
    mail_service_mock.send_mail.return_value = mail

    with container.override(MailService, instance=mail_service_mock):
        response = client.post("/send-mail", json={"message": mail.message})

    assert response.status_code == 200
    assert response.json() == {
        "email": mail.email,
        "message": mail.message,
    }


def test_send_mail_mock_user_service(
    client: TestClient, container: TestContainer
) -> None:
    user = User(id=999, email="mock1@mail.com")
    message = "hello"

    user_service_mock = mock.MagicMock(spec=UserService)
    user_service_mock.get_user.return_value = user

    with container.override(UserService, instance=user_service_mock):
        response = client.post("/send-mail", json={"message": message})

    assert response.status_code == 200
    assert response.json() == {
        "email": user.email,
        "message": message,
    }



from django.urls import include, path

from .views import (
    get_configured_dependency,
    get_request_scoped_dependency,
    get_request_scoped_dependency_async,
    get_setting,
    get_setting_async,
)

urlpatterns = [
    path("get-setting/", get_setting),
    path("get-setting-async/", get_setting_async),
    path("get-configured-dependency/", get_configured_dependency),
    path("get-request-scoped-dependency/", get_request_scoped_dependency),
    path("get-request-scoped-dependency-async/", get_request_scoped_dependency_async),
    path("api/", include("tests.ext.django.api.urls")),
]



import asyncio
import sys
import threading
import uuid
from collections.abc import AsyncIterator, Iterator, Sequence
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Union

import pytest
from typing_extensions import Self

from anydi import Container, Provider, Scope, auto, request, singleton, transient
from anydi._provider import ProviderKind
from anydi._typing import Event

from tests.fixtures import (
    Class,
    Resource,
    Service,
    UniqueId,
    async_event,
    async_generator,
    coro,
    event,
    func,
    generator,
    iterator,
)


@pytest.fixture
def container() -> Container:
    return Container()


class TestContainer:
    def test_register_provider(self, container: Container) -> None:
        def provider_call() -> str:
            return "test"

        provider = container.register(str, provider_call, scope="transient")

        assert provider.call == provider_call
        assert provider.scope == "transient"

    def test_provider_decorator(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def ident() -> str:
            return "1000"

        provider = container.providers[str]

        assert provider.call == ident
        assert provider.scope == "singleton"
        assert provider.interface is str

    @pytest.mark.parametrize(
        ("call", "kind", "interface"),
        [
            (func, ProviderKind.FUNCTION, str),
            (Class, ProviderKind.CLASS, Class),
            (generator, ProviderKind.GENERATOR, str),
            (async_generator, ProviderKind.ASYNC_GENERATOR, str),
            (coro, ProviderKind.COROUTINE, str),
        ],
    )
    def test_register_provider_different_kind(
        self,
        container: Container,
        call: Callable[..., Any],
        kind: ProviderKind,
        interface: Any,
    ) -> None:
        provider = container._register_provider(call, "singleton")

        assert provider.kind == kind
        assert provider.interface is interface

    @pytest.mark.parametrize(
        ("annotation", "expected"),
        [
            (str, str),
            (int, int),
            (Service, Service),
            (Iterator[Service], Service),
            (AsyncIterator[Service], Service),
            (dict[str, Any], dict[str, Any]),
            (list[str], list[str]),
            ("list[str]", list[str]),
            (tuple[str, ...], tuple[str, ...]),
            ("tuple[str, ...]", tuple[str, ...]),
            ('Annotated[str, "name"]', Annotated[str, "name"]),
        ],
    )
    def test_register_provider_interface(
        self, container: Container, annotation: type[Any], expected: type[Any]
    ) -> None:
        def call() -> annotation:  # type: ignore[valid-type]
            return object()

        provider = container._register_provider(call, "singleton")

        assert provider.interface == expected

    @pytest.mark.parametrize(
        ("call", "kind"),
        [
            (event, ProviderKind.GENERATOR),
            (async_event, ProviderKind.ASYNC_GENERATOR),
        ],
    )
    def test_register_provider_event(
        self,
        container: Container,
        call: Callable[..., Any],
        kind: ProviderKind,
    ) -> None:
        provider = container._register_provider(call, "singleton")

        assert provider.kind == kind
        assert issubclass(provider.interface, Event)

    def test_register_provider_with_interface(self, container: Container) -> None:
        provider = container._register_provider(lambda: "hello", "singleton", str)

        assert provider.interface is str

    def test_register_provider_with_none(self, container: Container) -> None:
        with pytest.raises(
            TypeError, match="Missing `(.*?)` provider return annotation."
        ):
            container._register_provider(lambda: "hello", "singleton", None)

    def test_register_provider_provider_without_return_annotation(
        self, container: Container
    ) -> None:
        def provide_message():  # type: ignore[no-untyped-def]
            return "hello"

        with pytest.raises(
            TypeError, match="Missing `(.*?)` provider return annotation."
        ):
            container._register_provider(provide_message, "singleton")

    def test_register_provider_not_callable(self, container: Container) -> None:
        with pytest.raises(
            TypeError,
            match="The provider `Test` is invalid because it is not a callable object.",
        ):
            container._register_provider("Test", "singleton")  # type: ignore[arg-type]

    def test_register_provider_iterator_no_arg_not_allowed(
        self, container: Container
    ) -> None:
        with pytest.raises(
            TypeError,
            match=(
                "Cannot use `(.*?)` resource type annotation "
                "without actual type argument."
            ),
        ):
            container._register_provider(iterator, "singleton")

    def test_register_provider_unsupported_scope(self, container: Container) -> None:
        with pytest.raises(
            ValueError,
            match=(
                "The provider `(.*?)` scope is invalid. Only the following scopes "
                "are supported: transient, singleton, request. "
                "Please use one of the supported scopes when registering a provider."
            ),
        ):
            container._register_provider(generator, "other")  # type: ignore[arg-type]

    def test_register_provider_transient_resource_not_allowed(
        self, container: Container
    ) -> None:
        with pytest.raises(
            TypeError,
            match=(
                "The resource provider `(.*?)` is attempting to "
                "register with a transient scope, which is not allowed."
            ),
        ):
            container._register_provider(generator, "transient")

    def test_register_provider_without_annotation(self, container: Container) -> None:
        def service_ident() -> str:
            return "10000"

        def service(ident) -> Service:  # type: ignore[no-untyped-def]
            return Service(ident=ident)

        with pytest.raises(
            TypeError, match="Missing provider `(.*?)` dependency `ident` annotation."
        ):
            container._register_provider(service, "singleton")

    def test_register_provider_positional_only_parameter_not_allowed(
        self, container: Container
    ) -> None:
        def provider_message(a: int, /, b: str) -> str:
            return f"{a} {b}"

        with pytest.raises(
            TypeError,
            match="Positional-only parameters are not allowed in the provider `(.*?)`.",
        ):
            container._register_provider(provider_message, "singleton")

    def test_register_provider_already_registered(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="singleton")

        with pytest.raises(
            LookupError, match="The provider interface `str` already registered."
        ):
            container.register(str, lambda: "other", scope="singleton")

    def test_register_provider_override(self, container: Container) -> None:
        container.register(str, lambda: "old", scope="singleton")

        def new_provider_call() -> str:
            return "new"

        provider = container.register(
            str, new_provider_call, scope="singleton", override=True
        )

        assert provider.call == new_provider_call

    def test_register_provider_named(self, container: Container) -> None:
        container.register(
            Annotated[str, "msg1"],
            lambda: "test1",
            scope="singleton",
        )
        container.register(
            Annotated[str, "msg2"],
            lambda: "test2",
            scope="singleton",
        )

        assert container.is_registered(Annotated[str, "msg1"])
        assert container.is_registered(Annotated[str, "msg2"])

    def test_register_providers_via_constructor(self) -> None:
        container = Container(
            providers=[
                Provider(call=lambda: "test", scope="singleton", interface=str),
                Provider(call=lambda: 1, scope="singleton", interface=int),
            ]
        )

        assert container.is_registered(str)
        assert container.is_registered(int)

    def test_register_provider_invalid_transient_resource(
        self, container: Container
    ) -> None:
        def provider_call() -> Iterator[str]:
            yield "test"

        with pytest.raises(
            TypeError,
            match=(
                "The resource provider `(.*?)` is attempting to register "
                "with a transient scope, which is not allowed."
            ),
        ):
            container.register(str, provider_call, scope="transient")

    def test_register_provider_invalid_transient_async_resource(
        self,
        container: Container,
    ) -> None:
        async def provider_call() -> AsyncIterator[str]:
            yield "test"

        with pytest.raises(
            TypeError,
            match=(
                "The resource provider `(.*?)` is attempting to register with a "
                "transient scope, which is not allowed."
            ),
        ):
            container.register(str, provider_call, scope="transient")

    def test_register_provider_valid_resource(self, container: Container) -> None:
        def provider_call1() -> Iterator[str]:
            yield "test"

        def provider_call2() -> Iterator[int]:
            yield 100

        container.register(str, provider_call1, scope="singleton")
        container.register(int, provider_call2, scope="request")

    def test_register_provider_valid_async_resource(self, container: Container) -> None:
        async def provider_call1() -> AsyncIterator[str]:
            yield "test"

        async def provider_call2() -> AsyncIterator[int]:
            yield 100

        container.register(str, provider_call1, scope="singleton")
        container.register(int, provider_call2, scope="request")

    def test_register_invalid_provider_type(self, container: Container) -> None:
        with pytest.raises(
            TypeError,
            match="The provider `Test` is invalid because it is not a callable object.",
        ):
            container.register(str, "Test", scope="singleton")  # type: ignore[arg-type]

    @pytest.mark.parametrize(
        ("scope1", "scope2", "scope3", "valid"),
        [
            ("transient", "transient", "transient", True),
            ("transient", "transient", "singleton", True),
            ("transient", "transient", "request", True),
            ("transient", "singleton", "transient", False),
            ("transient", "singleton", "singleton", True),
            ("transient", "singleton", "request", False),
            ("transient", "request", "transient", False),
            ("transient", "request", "singleton", True),
            ("transient", "request", "request", True),
            ("singleton", "transient", "transient", False),
            ("singleton", "transient", "singleton", False),
            ("singleton", "transient", "request", False),
            ("singleton", "singleton", "transient", False),
            ("singleton", "singleton", "singleton", True),
            ("singleton", "singleton", "request", False),
            ("singleton", "request", "transient", False),
            ("singleton", "request", "singleton", False),
            ("singleton", "request", "request", False),
            ("request", "transient", "transient", False),
            ("request", "transient", "singleton", False),
            ("request", "transient", "request", False),
            ("request", "singleton", "transient", False),
            ("request", "singleton", "singleton", True),
            ("request", "singleton", "request", False),
            ("request", "request", "transient", False),
            ("request", "request", "singleton", True),
            ("request", "request", "request", True),
        ],
    )
    def test_register_provider_match_scopes(
        self,
        container: Container,
        scope1: Scope,
        scope2: Scope,
        scope3: Scope,
        valid: bool,
    ) -> None:
        def a() -> int:
            return 2

        def b(a: int) -> float:
            return a * 2.5

        def mixed(a: int, b: float) -> str:
            return f"{a} * {b} = {a * b}"

        try:
            container.register(int, a, scope=scope3)
            container.register(float, b, scope=scope2)
            container.register(str, mixed, scope=scope1)
        except ValueError:
            result = False
        else:
            result = True

        assert result == valid

    def test_register_provider_match_scopes_error(self, container: Container) -> None:
        def provider_int() -> int:
            return 1000

        def provider_str(n: int) -> str:
            return f"{n}"

        container.register(int, provider_int, scope="request")

        with pytest.raises(
            ValueError,
            match=(
                "The provider `(.*?)` with a `singleton` scope cannot depend on "
                "`(.*?)` with a `request` scope. Please ensure all providers are "
                "registered with matching scopes."
            ),
        ):
            container.register(str, provider_str, scope="singleton")

    def test_register_provider_with_not_registered_sub_provider(
        self,
        container: Container,
    ) -> None:
        def dep2(dep1: int) -> str:
            return str(dep1)

        with pytest.raises(
            LookupError,
            match=(
                "The provider `(.*?).dep2` depends on `dep1` of type `int`, "
                "which has not been registered or set. To resolve this, "
                "ensure that `dep1` is registered before attempting to use it."
            ),
        ):
            container.register(str, dep2, scope="singleton")

    def test_register_events(self, container: Container) -> None:
        events = []

        @container.provider(scope="singleton")
        def message() -> str:
            return "test"

        @container.provider(scope="singleton")
        def event_1(message: str) -> Iterator[None]:
            events.append(f"event_1: before {message}")
            yield
            events.append(f"event_1: after {message}")

        @container.provider(scope="request")
        def event_2(message: str) -> Iterator[None]:
            events.append(f"event_2: before {message}")
            yield
            events.append(f"event_2: after {message}")

        # Ensure that non-event is not called
        @container.provider(scope="request")
        def non_event(message: str) -> Iterator[int]:
            events.append(f"non_event: before {message}")
            yield 1
            events.append(f"non_event: after {message}")

        with container, container.request_context():
            assert events == [
                "event_1: before test",
                "event_2: before test",
            ]

        assert events == [
            "event_1: before test",
            "event_2: before test",
            "event_2: after test",
            "event_1: after test",
        ]

    async def test_register_async_events(self, container: Container) -> None:
        events = []

        @container.provider(scope="singleton")
        def message() -> str:
            return "test"

        @container.provider(scope="singleton")
        async def event_1(message: str) -> AsyncIterator[None]:
            events.append(f"event_1: before {message}")
            yield
            events.append(f"event_1: after {message}")

        @container.provider(scope="request")
        def event_2(message: str) -> Iterator[None]:
            events.append(f"event_2: before {message}")
            yield
            events.append(f"event_2: after {message}")

        # Ensure that non-event is not called
        @container.provider(scope="request")
        async def non_event(message: str) -> AsyncIterator[int]:
            events.append(f"non_event: before {message}")
            yield 1
            events.append(f"non_event: after {message}")

        async with container, container.arequest_context():
            assert events == [
                "event_1: before test",
                "event_2: before test",
            ]

        assert events == [
            "event_1: before test",
            "event_2: before test",
            "event_2: after test",
            "event_1: after test",
        ]

    def test_unregister_provider(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="singleton")

        assert container.is_registered(str)

        container.unregister(str)

        assert not container.is_registered(str)

    def test_unregister_provider_resource(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> Iterator[str]:
            yield "test"

        assert container.is_registered(str)
        assert str in container._resources["singleton"]

        container.unregister(str)

        assert not container.is_registered(str)
        assert str not in container._resources["singleton"]

    def test_unregister_request_scoped_provider(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="request")

        assert container.is_registered(str)

        container.unregister(str)

        assert not container.is_registered(str)

    def test_unregister_not_registered_provider(self, container: Container) -> None:
        with pytest.raises(
            LookupError, match="The provider interface `str` not registered."
        ):
            container.unregister(str)

    # Lifespan

    def test_start_and_close_singleton_context(self, container: Container) -> None:
        events = []

        def dep1() -> Iterator[str]:
            events.append("dep1:before")
            yield "test"
            events.append("dep1:after")

        container.register(str, dep1, scope="singleton")

        container.start()

        assert container.resolve(str) == "test"

        container.close()

        assert events == ["dep1:before", "dep1:after"]

    def test_request_context(self, container: Container) -> None:
        events = []

        def dep1() -> Iterator[str]:
            events.append("dep1:before")
            yield "test"
            events.append("dep1:after")

        container.register(str, dep1, scope="request")

        with container.request_context():
            assert container.resolve(str) == "test"

        assert events == ["dep1:before", "dep1:after"]

    # Asynchronous lifespan

    async def test_astart_and_aclose_singleton_context(
        self, container: Container
    ) -> None:
        events = []

        async def dep1() -> AsyncIterator[str]:
            events.append("dep1:before")
            yield "test"
            events.append("dep1:after")

        container.register(str, dep1, scope="singleton")

        await container.astart()

        assert container.resolve(str) == "test"

        await container.aclose()

        assert events == ["dep1:before", "dep1:after"]

    async def test_arequest_context(self, container: Container) -> None:
        events = []

        async def dep1() -> AsyncIterator[str]:
            events.append("dep1:before")
            yield "test"
            events.append("dep1:after")

        container.register(str, dep1, scope="request")

        async with container.arequest_context():
            assert await container.aresolve(str) == "test"

        assert events == ["dep1:before", "dep1:after"]

    def test_reset_resolved_instances(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="singleton")
        container.register(int, lambda: 1, scope="singleton")

        container.resolve(str)
        container.resolve(int)

        assert container.is_resolved(str)
        assert container.is_resolved(int)

        container.reset()

        assert not container.is_resolved(str)
        assert not container.is_resolved(int)

    def test_reset_transient(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="transient")

        _ = container.resolve(str)

        assert not container.is_resolved(str)

        container.reset()

        assert not container.is_resolved(str)

    # Instance

    def test_resolve_singleton_scoped(self, container: Container) -> None:
        instance = "test"

        container.register(str, lambda: instance, scope="singleton")

        assert container.resolve(str) == instance

    def test_resolve_singleton_scoped_not_started(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> Iterator[str]:
            yield "test"

        assert container.resolve(str) == "test"

    def test_resolve_singleton_scoped_resource(self, container: Container) -> None:
        instance = "test"

        def provide() -> Iterator[str]:
            yield instance

        container.register(str, provide, scope="singleton")
        container.start()

        assert container.resolve(str) == instance

    def test_resolve_singleton_scoped_started_with_async_resource_provider(
        self,
        container: Container,
    ) -> None:
        instance = "test"

        async def provide() -> AsyncIterator[str]:
            yield instance

        container.register(str, provide, scope="singleton")

        with pytest.raises(
            TypeError,
            match=(
                "The instance for the provider `(.*?).provide` cannot be created "
                "in synchronous mode."
            ),
        ):
            container.start()

    def test_resolve_singleton_resource(self, container: Container) -> None:
        instance = "test"

        def provide() -> Iterator[str]:
            yield instance

        container.register(str, provide, scope="singleton")

        container.resolve(str)

    def test_resolve_singleton_scoped_thread_safe(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def provide_unique_id() -> UniqueId:
            return UniqueId()

        unique_ids = set()

        def use_unique_id() -> None:
            unique_id = container.resolve(UniqueId)
            unique_ids.add(unique_id)

        threads = [threading.Thread(target=use_unique_id) for _ in range(10)]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert len(unique_ids) == 1

    async def test_resolve_singleton_async_resource(self, container: Container) -> None:
        instance = "test"

        def provide() -> Iterator[str]:
            yield instance

        container.register(str, provide, scope="singleton")

        await container.astart()

        assert container.resolve(str) == instance

    async def test_resolve_singleton_async_and_sync_resources(
        self, container: Container
    ) -> None:
        instance_str = "test"
        instance_int = 100

        def provider_1() -> Iterator[str]:
            yield instance_str

        async def provider_2() -> AsyncIterator[int]:
            yield instance_int

        container.register(str, provider_1, scope="singleton")
        container.register(int, provider_2, scope="singleton")

        await container.astart()

        assert container.resolve(str) == instance_str
        assert container.resolve(int) == instance_int

    async def test_resolved_singleton_async_resource_not_started(
        self,
        container: Container,
    ) -> None:
        instance = "test"

        async def provide() -> AsyncIterator[str]:
            yield instance

        container.register(str, provide, scope="singleton")

        with pytest.raises(
            TypeError,
            match=(
                "The instance for the provider `(.*?).provide` cannot be created "
                "in synchronous mode."
            ),
        ):
            container.resolve(str)

    def test_resolve_singleton_annotated_resource(self, container: Container) -> None:
        instance = "test"

        @container.provider(scope="singleton")
        def provide() -> Iterator[Annotated[str, "message"]]:
            yield instance

        result = container.resolve(Annotated[str, "message"])

        assert result == instance

    async def test_resolve_singleton_annotated_async_resource(
        self, container: Container
    ) -> None:
        instance = "test"

        @container.provider(scope="singleton")
        async def provide() -> AsyncIterator[Annotated[str, "message"]]:
            yield instance

        result = await container.aresolve(Annotated[str, "message"])

        assert result == instance

    async def test_resolve_singleton_scoped_coro_safe(
        self, container: Container
    ) -> None:
        @container.provider(scope="singleton")
        async def provide_unique_id() -> UniqueId:
            return UniqueId()

        unique_ids = set()

        async def use_service() -> None:
            unique_id = await container.aresolve(UniqueId)
            unique_ids.add(unique_id)

        tasks = [use_service() for _ in range(10)]

        await asyncio.gather(*tasks)

        assert len(unique_ids) == 1

    async def test_resolve_scoped_coro_safe(self, container: Container) -> None:
        @container.provider(scope="request")
        async def provide_unique_id() -> UniqueId:
            return UniqueId()

        unique_ids = set()

        async def use_unique_id() -> None:
            async with container.arequest_context():
                unique_id = await container.aresolve(UniqueId)
                unique_ids.add(unique_id)

        tasks = [use_unique_id() for _ in range(10)]
        await asyncio.gather(*tasks)

        assert len(unique_ids) == 10

    def test_resolve_request_scoped(self, container: Container) -> None:
        instance = "test"

        container.register(str, lambda: instance, scope="request")

        with container.request_context():
            assert container.resolve(str) == instance

    def test_resolve_request_scoped_not_started(self, container: Container) -> None:
        instance = "test"

        container.register(str, lambda: instance, scope="request")

        with pytest.raises(
            LookupError,
            match=(
                "The request context has not been started. Please ensure that the "
                "request context is properly initialized before attempting to use it."
            ),
        ):
            container.resolve(str)

    def test_resolve_request_scoped_annotated_resource(
        self, container: Container
    ) -> None:
        instance = "test"

        @container.provider(scope="request")
        def provide() -> Iterator[Annotated[str, "message"]]:
            yield instance

        with container.request_context():
            result = container.resolve(Annotated[str, "message"])

        assert result == instance

    async def test_resolve_request_scoped_annotated_async_resource(
        self,
        container: Container,
    ) -> None:
        instance = "test"

        @container.provider(scope="request")
        async def provide() -> AsyncIterator[Annotated[str, "message"]]:
            yield instance

        async with container.arequest_context():
            result = await container.aresolve(Annotated[str, "message"])

        assert result == instance

    def test_resolve_request_scoped_unresolved_yet(self, container: Container) -> None:
        class Request:
            def __init__(self, path: str) -> None:
                self.path = path

        @container.provider(scope="request")
        def req_path(req: Request) -> str:
            return req.path

        with container.request_context() as context:
            context.set(Request, Request(path="test"))
            assert container.resolve(str) == "test"

    def test_resolve_request_scoped_unresolved_error(
        self, container: Container
    ) -> None:
        class Request:
            def __init__(self, path: str) -> None:
                self.path = path

        @container.provider(scope="request")
        def req_path(req: Request) -> str:
            return req.path

        with (
            pytest.raises(
                LookupError,
                match=(
                    "You are attempting to get the parameter `req` with the annotation "
                    "`(.*?).Request` as a dependency into `(.*?).req_path` which is "
                    "not registered or set in the scoped context."
                ),
            ),
            container.request_context(),
        ):
            container.resolve(str)

    def test_resolve_scoped_thread_safe(self, container: Container) -> None:
        @container.provider(scope="request")
        def provide_unique_id() -> UniqueId:
            return UniqueId()

        unique_ids = set()

        def use_unique_id() -> None:
            with container.request_context():
                unique_id = container.resolve(UniqueId)
                unique_ids.add(unique_id)

        threads = [threading.Thread(target=use_unique_id) for n in range(10)]

        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        assert len(unique_ids) == 10

    def test_resolve_transient_scoped(self, container: Container) -> None:
        container.register(uuid.UUID, uuid.uuid4, scope="transient")

        assert container.resolve(uuid.UUID) != container.resolve(uuid.UUID)

    def test_sync_resolve_transient_async_provider(self, container: Container) -> None:
        @container.provider(scope="transient")
        async def get_uuid() -> uuid.UUID:
            return uuid.uuid4()

        with pytest.raises(
            TypeError,
            match=(
                "The instance for the provider `(.*?).get_uuid` cannot "
                "be created in synchronous mode."
            ),
        ):
            container.resolve(uuid.UUID)

    async def test_async_resolve_transient_provider(self, container: Container) -> None:
        @container.provider(scope="transient")
        async def get_uuid() -> uuid.UUID:
            return uuid.uuid4()

        assert await container.aresolve(uuid.UUID) != await container.aresolve(
            uuid.UUID
        )

    async def test_async_resolve_synchronous_resource(
        self, container: Container
    ) -> None:
        @container.provider(scope="singleton")
        def msg() -> Iterator[str]:
            yield "test"

        assert await container.aresolve(str) == "test"

    def test_resolve_not_registered_instance(self, container: Container) -> None:
        with pytest.raises(
            LookupError,
            match=(
                "The provider interface `str` is either not registered, not provided, "
                "or not set in the scoped context. Please ensure that the provider "
                "interface is properly registered and that the class is decorated "
                "with a scope before attempting to use it."
            ),
        ):
            container.resolve(str)

    def test_resolve_non_strict_provider_scope_defined(
        self, container: Container
    ) -> None:
        @singleton
        class Service:
            pass

        _ = container.resolve(Service)

        provider = container.providers[Service]

        assert provider.call == Service
        assert provider.scope == "singleton"
        assert provider.interface == Service

    def test_resolve_non_strict_provider_scope_from_sub_provider_request(
        self,
        container: Container,
    ) -> None:
        @container.provider(scope="request")
        def ident() -> str:
            return "test"

        with container.request_context():
            _ = container.resolve(Service)

        assert Service in container.providers

        provider = container.providers[str]

        assert provider.call == ident
        assert provider.scope == "request"
        assert provider.interface is str

    def test_resolve_non_strict_provider_scope_from_sub_provider_transient(
        self,
        container: Container,
    ) -> None:
        @container.provider(scope="transient")
        def uuid_generator() -> Annotated[str, "uuid_generator"]:
            return str(uuid.uuid4())

        @dataclass
        class Entity:
            id: Annotated[str, "uuid_generator"]

        _ = container.resolve(Entity)

        assert container.providers[Entity].scope == "transient"

    def test_resolve_non_strict_nested_singleton_provider(
        self, container: Container
    ) -> None:
        @singleton
        class Repository:
            pass

        class Service:
            def __init__(self, repository: Repository) -> None:
                self.repository = repository

        with container.request_context():
            _ = container.resolve(Service)

        assert container.providers[Service].scope == "singleton"

    def test_resolve_non_strict_default_scope(self, container: Container) -> None:
        @dataclass
        class Repository:
            pass

        @dataclass
        class Service:
            repository: Repository

        _ = container.resolve(Service)

        assert container.providers[Service].scope == "transient"

    def test_resolve_non_strict_with_primitive_class(
        self, container: Container
    ) -> None:
        with pytest.raises(
            LookupError,
            match=(
                "The provider `tests.fixtures.Service` depends on `ident` of type "
                "`str`, which has not been registered or set. To resolve this, "
                "ensure that `ident` is registered before attempting to use it."
            ),
        ):
            _ = container.resolve(Service).ident

    @pytest.mark.skipif(sys.version_info < (3, 10), reason="Requires Python 3.10")
    def test_resolve_non_strict_with_custom_type(self, container: Container) -> None:
        class Klass:
            def __init__(
                self, value: "Union[str, Sequence[str], int, list[str]]"
            ) -> None:
                self.value = value

        with pytest.raises(
            LookupError,
            match=(
                "The provider `(.*?)` depends on `value` of type `(.*?)`, "
                "which has not been registered or set. To resolve this, "
                "ensure that `value` is registered before attempting to use it."
            ),
        ):
            _ = container.resolve(Klass)

    def test_resolve_non_strict_with_as_context_manager(
        self, container: Container
    ) -> None:
        @singleton
        class Resource:
            def __init__(self) -> None:
                self.entered = False
                self.exited = False

            def __enter__(self) -> Self:
                self.entered = True
                return self

            def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
                self.exited = True

        resource = container.resolve(Resource)

        assert resource.entered

        container.close()

        assert resource.exited

    async def test_resolve_non_strict_with_as_async_context_manager(
        self,
        container: Container,
    ) -> None:
        @singleton
        class Service:
            def __init__(self) -> None:
                self.entered = False
                self.exited = False

            async def __aenter__(self) -> Self:
                self.entered = True
                return self

            async def __aexit__(
                self, exc_type: Any, exc_value: Any, traceback: Any
            ) -> None:
                self.exited = True

        service = await container.aresolve(Service)

        assert service.entered

        await container.aclose()

        assert service.exited

    def test_resolve_non_strict_with_defaults(self, container: Container) -> None:
        @dataclass
        class Repo:
            name: str = "repo"

        class Service:
            def __init__(self, repo: Repo, name: str = "service") -> None:
                self.repo = repo
                self.name = name

        service = container.resolve(Service)

        assert service.name == "service"
        assert service.repo.name == "repo"

    async def test_resolve_non_strict_with_defaults_async_resolver(
        self,
        container: Container,
    ) -> None:
        @dataclass
        class Repo:
            name: str = "repo"

        class Service:
            def __init__(self, repo: Repo, name: str = "service") -> None:
                self.repo = repo
                self.name = name

        service = await container.aresolve(Service)

        assert service.name == "service"
        assert service.repo.name == "repo"

    def test_is_resolved(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> str:
            return "test"

        _ = container.resolve(str)

        assert container.is_resolved(str)

    def test_is_resolved_transient(self, container: Container) -> None:
        @container.provider(scope="transient")
        def message() -> str:
            return "test"

        _ = container.resolve(str)

        assert not container.is_resolved(str)

    def test_is_resolved_false(self, container: Container) -> None:
        assert not container.is_resolved(str)

    def test_release_instance(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="singleton")
        container.resolve(str)

        assert container.is_resolved(str)

        container.release(str)

        assert not container.is_resolved(str)

    def test_release_transient_instance(self, container: Container) -> None:
        container.register(str, lambda: "test", scope="transient")

        assert not container.is_resolved(str)

        container.release(str)

        assert not container.is_resolved(str)

    def test_resource_delegated_exception(self, container: Container) -> None:
        resource = Resource()

        @container.provider(scope="request")
        def resource_provider() -> Iterator[Resource]:
            try:
                yield resource
            except Exception:  # noqa
                resource.rollback()
                raise
            else:
                resource.commit()

        def _resolve() -> None:
            resource = container.resolve(Resource)
            resource.run()
            raise ValueError("error")

        with pytest.raises(ValueError, match="error"), container.request_context():
            _resolve()

        assert resource.called
        assert not resource.committed
        assert resource.rolled_back

    async def test_async_resource_delegated_exception(
        self, container: Container
    ) -> None:
        resource = Resource()

        @container.provider(scope="request")
        async def resource_provider() -> AsyncIterator[Resource]:
            try:
                yield resource
            except Exception:  # noqa
                resource.rollback()
                raise
            else:
                resource.commit()

        async def _resolve() -> None:
            resource = await container.aresolve(Resource)
            resource.run()
            raise ValueError("error")

        with pytest.raises(ValueError, match="error"):
            async with container.arequest_context():
                await _resolve()

        assert resource.called
        assert not resource.committed
        assert resource.rolled_back

    # Inspections

    def test_get_provider_arguments(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def a() -> int:
            return 10

        @container.provider(scope="singleton")
        def b() -> float:
            return 1.0

        @container.provider(scope="singleton")
        def c() -> str:
            return "test"

        def service(a: int, b: float, *, c: str) -> Service:
            return Service(ident=f"{a}/{b}/{c}")

        provider = container.register(Service, service, scope="singleton")

        context = container._get_instance_context("singleton")

        kwargs = container._get_provided_kwargs(provider, context)

        assert kwargs == {"a": 10, "b": 1.0, "c": "test"}

    async def test_async_get_provider_arguments(self, container: Container) -> None:
        @container.provider(scope="singleton")
        async def a() -> int:
            return 10

        @container.provider(scope="singleton")
        async def b() -> float:
            return 1.0

        @container.provider(scope="singleton")
        async def c() -> str:
            return "test"

        async def service(a: int, b: float, *, c: str) -> Service:
            return Service(ident=f"{a}/{b}/{c}")

        provider = container.register(Service, service, scope="singleton")

        context = container._get_instance_context("singleton")

        kwargs = await container._aget_provided_kwargs(provider, context)

        assert kwargs == {"a": 10, "b": 1.0, "c": "test"}

    def test_create_transient_non_strict(self) -> None:
        @transient
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        instance = container.create(Component, name="test")

        assert instance.name == "test"

    def test_create_singleton_non_strict(self) -> None:
        @singleton
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        instance = container.create(Component, name="test")

        assert instance.name == "test"

    def test_create_scoped_non_strict(self) -> None:
        @request
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        with container.request_context():
            instance = container.create(Component, name="test")

        assert instance.name == "test"

    def test_create_non_existing_keyword_arg(self) -> None:
        @singleton
        class Component:
            pass

        container = Container()

        with pytest.raises(TypeError, match="takes no arguments"):
            container.create(Component, param="test")

    async def test_create_async_transient_non_strict(self) -> None:
        @transient
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        instance = await container.acreate(Component, name="test")

        assert instance.name == "test"

    async def test_create_async_singleton_non_strict(self) -> None:
        @singleton
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        instance = await container.acreate(Component, name="test")

        assert instance.name == "test"

    async def test_create_async_scoped_non_strict(self) -> None:
        @request
        class Component:
            def __init__(self, name: str) -> None:
                self.name = name

        container = Container()

        with container.request_context():
            instance = await container.acreate(Component, name="test")

        assert instance.name == "test"

    async def test_create_async_non_existing_keyword_arg(self) -> None:
        @singleton
        class Component:
            pass

        container = Container()

        with pytest.raises(TypeError, match="takes no arguments"):
            await container.acreate(Component, param="test")


class TestContainerInjector:
    def test_inject(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def ident_provider() -> str:
            return "1000"

        @container.provider(scope="singleton")
        def service_provider(ident: str) -> Service:
            return Service(ident=ident)

        @container.inject
        def func(name: str, service: Service = auto) -> str:
            return f"{name} = {service.ident}"

        result = func(name="service ident")

        assert result == "service ident = 1000"

    @pytest.mark.skip(reason="disable until strict is enforced")
    def test_inject_missing_annotation(self, container: Container) -> None:
        def handler(name=auto) -> str:  # type: ignore
            return name  # type: ignore

        with pytest.raises(
            TypeError, match="Missing `(.*?).handler` parameter `name` annotation."
        ):
            container.inject(handler)

    @pytest.mark.skip(reason="disable until strict is enforced")
    def test_inject_unknown_dependency_using_strict_mode(self) -> None:
        container = Container()

        def handler(message: str = auto) -> None:
            pass

        with pytest.raises(
            LookupError,
            match=(
                "`(.*?).handler` has an unknown dependency parameter `message` "
                "with an annotation of `str`."
            ),
        ):
            container.inject(handler)

    def test_inject_auto_marker(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> str:
            return "test"

        @container.inject
        def func(message: str = auto) -> str:
            return message

        result = func()

        assert result == "test"

    def test_inject_auto_marker_call(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> str:
            return "test"

        @container.inject
        def func(message: str = auto()) -> str:
            return message

        result = func()

        assert result == "test"

    def test_inject_class(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def ident_provider() -> str:
            return "1000"

        @container.provider(scope="singleton")
        def service_provider(ident: str) -> Service:
            return Service(ident=ident)

        @container.inject
        class Handler:
            def __init__(self, name: str, service: Service = auto) -> None:
                self.name = name
                self.service = service

            def handle(self) -> str:
                return f"{self.name} = {self.service.ident}"

        handler = Handler(name="service ident")

        result = handler.handle()

        assert result == "service ident = 1000"

    async def test_inject_with_sync_and_async_resources(
        self, container: Container
    ) -> None:
        def ident_provider() -> Iterator[str]:
            yield "1000"

        async def service_provider(ident: str) -> AsyncIterator[Service]:
            yield Service(ident=ident)

        container.register(str, ident_provider, scope="singleton")
        container.register(Service, service_provider, scope="singleton")

        await container.astart()

        @container.inject
        async def func(name: str, service: Service = auto) -> str:
            return f"{name} = {service.ident}"

        result = await func(name="service ident")

        assert result == "service ident = 1000"

    def test_run(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def value1() -> Annotated[int, "value1"]:
            return 10

        @container.provider(scope="singleton")
        def value2() -> Annotated[int, "value2"]:
            return 20

        def sum_handler(
            value1: int,
            value2: Annotated[int, "value1"] = auto,
            value3: Annotated[int, "value2"] = auto,
        ) -> int:
            return value1 + value2 + value3

        result = container.run(sum_handler, value1=30)

        assert result == 60

    def test_run_cached(self, container: Container) -> None:
        @container.provider(scope="singleton")
        def message() -> str:
            return "hello"

        def handler(message: str = auto) -> str:
            return message

        _ = container.run(handler)
        _ = container.run(handler)

        assert handler in container._inject_cache



from anydi import auto, injectable

from tests.fixtures import Service


@injectable(tags=["inject"])
def a_a3_handler_1(message: str = auto) -> str:
    return message


def a_a3_handler_2(service: Service = auto) -> Service:
    return service



import uuid
from collections.abc import AsyncIterator, Iterator
from dataclasses import dataclass
from typing import Annotated
from unittest import mock

import pytest

from anydi import Container, auto, singleton
from anydi.testing import TestContainer

from tests.fixtures import Service


class TestTestContainer:
    def test_from_container(self) -> None:
        container = Container()
        container.register(str, lambda: "Hello, world!", scope="singleton")

        test_container = TestContainer.from_container(container)

        assert test_container.providers == container.providers
        assert test_container.default_scope == container.default_scope

        assert test_container.resolve(str) == "Hello, world!"

    def test_override_instance(self) -> None:
        origin_name = "origin"
        overridden_name = "overridden"

        container = TestContainer()

        @container.provider(scope="singleton")
        def name() -> str:
            return origin_name

        with container.override(str, overridden_name):
            assert container.resolve(str) == overridden_name

        assert container.resolve(str) == origin_name

    def test_override_instance_provider_not_registered_using_strict_mode(self) -> None:
        container = TestContainer()

        with pytest.raises(
            LookupError, match="The provider interface `str` not registered."
        ):
            with container.override(str, "test"):
                pass

    def test_override_instance_transient_provider(self) -> None:
        overridden_uuid = uuid.uuid4()

        container = TestContainer()

        @container.provider(scope="transient")
        def uuid_provider() -> uuid.UUID:
            return uuid.uuid4()

        with container.override(uuid.UUID, overridden_uuid):
            assert container.resolve(uuid.UUID) == overridden_uuid

        assert container.resolve(uuid.UUID) != overridden_uuid

    def test_override_instance_resource_provider(self) -> None:
        origin = "origin"
        overridden = "overridden"

        container = TestContainer()

        @container.provider(scope="singleton")
        def message() -> Iterator[str]:
            yield origin

        with container.override(str, overridden):
            assert container.resolve(str) == overridden

        assert container.resolve(str) == origin

    async def test_override_instance_async_resource_provider(self) -> None:
        origin = "origin"
        overridden = "overridden"

        container = TestContainer()

        @container.provider(scope="singleton")
        async def message() -> AsyncIterator[str]:
            yield origin

        with container.override(str, overridden):
            assert (await container.aresolve(str)) == overridden

    def test_override_registered_instance(self) -> None:
        container = TestContainer()
        container.register(Annotated[str, "param"], lambda: "param", scope="singleton")

        class UserRepo:
            def get_user(self) -> str:
                return "user"

        @singleton
        class UserService:
            def __init__(self, repo: UserRepo, param: Annotated[str, "param"]) -> None:
                self.repo = repo
                self.param = param

            def process(self) -> dict[str, str]:
                return {
                    "user": self.repo.get_user(),
                    "param": self.param,
                }

        user_repo_mock = mock.MagicMock(spec=UserRepo)
        user_repo_mock.get_user.return_value = "mocked_user"

        user_service = container.resolve(UserService)

        with (
            container.override(UserRepo, user_repo_mock),
            container.override(Annotated[str, "param"], "mock"),
        ):
            assert user_service.process() == {
                "user": "mocked_user",
                "param": "mock",
            }

    async def test_override_instance_async_resolved(self) -> None:
        container = TestContainer()
        container.register(Annotated[str, "param"], lambda: "param", scope="singleton")

        @singleton
        class UserService:
            def __init__(self, param: Annotated[str, "param"]) -> None:
                self.param = param

            def process(self) -> dict[str, str]:
                return {
                    "param": self.param,
                }

        user_service = await container.aresolve(UserService)

        with container.override(Annotated[str, "param"], "mock"):
            assert user_service.process() == {
                "param": "mock",
            }

    def test_override_instance_in_strict_mode(self) -> None:
        container = TestContainer()

        class Settings:
            def __init__(self, name: str) -> None:
                self.name = name

        @container.provider(scope="singleton")
        def provide_settings() -> Settings:
            return Settings(name="test")

        @container.provider(scope="singleton")
        def provide_service(settings: Settings) -> Service:
            return Service(ident=settings.name)

        service = container.resolve(Service)

        assert service.ident == "test"

    def test_override_instance_first(self) -> None:
        container = TestContainer()

        @dataclass
        class Item:
            name: str

        class ItemRepository:
            def __init__(self, items: list[Item]) -> None:
                self.items = items

            def all(self) -> list[Item]:
                return self.items

        class ItemService:
            def __init__(self, repo: ItemRepository) -> None:
                self.repo = repo

            def get_items(self) -> list[Item]:
                return self.repo.all()

        @container.provider(scope="singleton")
        def provide_repo() -> ItemRepository:
            return ItemRepository(items=[])

        @container.provider(scope="singleton")
        def provide_service(repo: ItemRepository) -> ItemService:
            return ItemService(repo=repo)

        @container.inject
        def handler(service: ItemService = auto) -> list[Item]:
            return service.get_items()

        repo_mock = mock.MagicMock(spec=ItemRepository)
        repo_mock.all.return_value = [Item(name="mocked")]

        with container.override(ItemRepository, repo_mock):
            items = handler()

            assert items == [Item(name="mocked")]

        service = container.resolve(ItemService)

        assert service.get_items() == []

    def test_override_prop(self) -> None:
        class ServiceWithProp:
            def __init__(self, name: str = "origin") -> None:
                self.name = name

        container = TestContainer()

        service = container.resolve(ServiceWithProp)

        assert service.name == "origin"



from __future__ import annotations

import uuid
from collections.abc import Iterator
from typing import Annotated

import anydi

from .services import HelloService


def configure(container: anydi.Container) -> None:
    container.register(
        Annotated[str, "configured-string"],
        lambda: "This is configured string",
        scope="singleton",
    )

    @container.provider(scope="singleton")
    def hello_service() -> Iterator[HelloService]:
        with HelloService() as hello_service:
            yield hello_service

    @container.provider(scope="request")
    def request_id() -> Annotated[str, "request-id"]:
        return uuid.uuid4().hex



from typing import cast

import pytest
from starlette.applications import Starlette
from starlette.testclient import TestClient

from anydi import Container

from .app import app as _app


@pytest.fixture(scope="session")
def app() -> Starlette:
    return _app


@pytest.fixture(scope="session")
def container(app: Starlette) -> Container:
    return cast(Container, app.state.container)


@pytest.fixture
def client(app: Starlette) -> TestClient:
    return TestClient(app)



from collections.abc import AsyncIterator

import pytest

from anydi.ext.django import container


@pytest.fixture(scope="session", autouse=True)
async def start_container() -> AsyncIterator[None]:
    await container.astart()
    yield
    await container.aclose()



"""AnyDI core implementation module."""

from __future__ import annotations

import contextlib
import functools
import inspect
import logging
import types
import uuid
from collections import defaultdict
from collections.abc import AsyncIterator, Iterable, Iterator
from contextvars import ContextVar
from typing import Any, Callable, TypeVar, cast, overload

from typing_extensions import ParamSpec, Self, get_args, get_origin

from ._async import run_sync
from ._context import InstanceContext
from ._decorators import is_provided
from ._module import ModuleDef, ModuleRegistrar
from ._provider import Provider, ProviderDef, ProviderKind
from ._scan import PackageOrIterable, Scanner
from ._scope import ALLOWED_SCOPES, Scope
from ._typing import (
    NOT_SET,
    Event,
    get_typed_annotation,
    get_typed_parameters,
    is_async_context_manager,
    is_builtin_type,
    is_context_manager,
    is_event_type,
    is_iterator_type,
    is_marker,
    is_none_type,
    type_repr,
)

T = TypeVar("T", bound=Any)
P = ParamSpec("P")


class Container:
    """AnyDI is a dependency injection container."""

    def __init__(
        self,
        *,
        providers: Iterable[ProviderDef] | None = None,
        modules: Iterable[ModuleDef] | None = None,
        default_scope: Scope = "transient",
        logger: logging.Logger | None = None,
    ) -> None:
        self._providers: dict[Any, Provider] = {}
        self._default_scope: Scope = default_scope
        self._logger = logger or logging.getLogger(__name__)
        self._resources: dict[str, list[Any]] = defaultdict(list)
        self._singleton_context = InstanceContext()
        self._request_context_var: ContextVar[InstanceContext | None] = ContextVar(
            "request_context", default=None
        )
        self._unresolved_interfaces: set[Any] = set()
        self._inject_cache: dict[Callable[..., Any], Callable[..., Any]] = {}

        # Components
        self._modules = ModuleRegistrar(self)
        self._scanner = Scanner(self)

        # Register providers
        providers = providers or []
        for provider in providers:
            self._register_provider(
                provider.call,
                provider.scope,
                provider.interface,
            )

        # Register modules
        modules = modules or []
        for module in modules:
            self.register_module(module)

    ############################
    # Properties
    ############################

    @property
    def default_scope(self) -> Scope:
        """Get the default scope."""
        return self._default_scope

    @property
    def providers(self) -> dict[type[Any], Provider]:
        """Get the registered providers."""
        return self._providers

    @property
    def logger(self) -> logging.Logger:
        """Get the logger instance."""
        return self._logger

    ############################
    # Lifespan/Context Methods
    ############################

    def __enter__(self) -> Self:
        """Enter the singleton context."""
        self.start()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> Any:
        """Exit the singleton context."""
        return self._singleton_context.__exit__(exc_type, exc_val, exc_tb)

    def start(self) -> None:
        """Start the singleton context."""
        # Resolve all singleton resources
        for interface in self._resources.get("singleton", []):
            self.resolve(interface)

    def close(self) -> None:
        """Close the singleton context."""
        self._singleton_context.close()

    async def __aenter__(self) -> Self:
        """Enter the singleton context."""
        await self.astart()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> bool:
        """Exit the singleton context."""
        return await self._singleton_context.__aexit__(exc_type, exc_val, exc_tb)

    async def astart(self) -> None:
        """Start the singleton context asynchronously."""
        for interface in self._resources.get("singleton", []):
            await self.aresolve(interface)

    async def aclose(self) -> None:
        """Close the singleton context asynchronously."""
        await self._singleton_context.aclose()

    @contextlib.contextmanager
    def request_context(self) -> Iterator[InstanceContext]:
        """Obtain a context manager for the request-scoped context."""
        context = InstanceContext()

        token = self._request_context_var.set(context)

        # Resolve all request resources
        for interface in self._resources.get("request", []):
            if not is_event_type(interface):
                continue
            self.resolve(interface)

        with context:
            yield context
            self._request_context_var.reset(token)

    @contextlib.asynccontextmanager
    async def arequest_context(self) -> AsyncIterator[InstanceContext]:
        """Obtain an async context manager for the request-scoped context."""
        context = InstanceContext()

        token = self._request_context_var.set(context)

        for interface in self._resources.get("request", []):
            if not is_event_type(interface):
                continue
            await self.aresolve(interface)

        async with context:
            yield context
            self._request_context_var.reset(token)

    def _get_request_context(self) -> InstanceContext:
        """Get the current request context."""
        request_context = self._request_context_var.get()
        if request_context is None:
            raise LookupError(
                "The request context has not been started. Please ensure that "
                "the request context is properly initialized before attempting "
                "to use it."
            )
        return request_context

    def _get_instance_context(self, scope: Scope) -> InstanceContext:
        """Get the instance context for the specified scope."""
        if scope == "singleton":
            return self._singleton_context
        return self._get_request_context()

    ############################
    # Provider Methods
    ############################

    def register(
        self,
        interface: Any,
        call: Callable[..., Any],
        *,
        scope: Scope,
        override: bool = False,
    ) -> Provider:
        """Register a provider for the specified interface."""
        return self._register_provider(call, scope, interface, override)

    def is_registered(self, interface: Any) -> bool:
        """Check if a provider is registered for the specified interface."""
        return interface in self._providers

    def has_provider_for(self, interface: Any) -> bool:
        """Check if a provider exists for the specified interface."""
        return self.is_registered(interface) or is_provided(interface)

    def unregister(self, interface: Any) -> None:
        """Unregister a provider by interface."""
        if not self.is_registered(interface):
            raise LookupError(
                f"The provider interface `{type_repr(interface)}` not registered."
            )

        provider = self._get_provider(interface)

        # Cleanup instance context
        if provider.scope != "transient":
            try:
                context = self._get_instance_context(provider.scope)
            except LookupError:
                pass
            else:
                del context[interface]

        # Cleanup provider references
        self._delete_provider(provider)

    def provider(
        self, *, scope: Scope, override: bool = False
    ) -> Callable[[Callable[P, T]], Callable[P, T]]:
        """Decorator to register a provider function with the specified scope."""

        def decorator(call: Callable[P, T]) -> Callable[P, T]:
            self._register_provider(call, scope, NOT_SET, override)
            return call

        return decorator

    def _register_provider(  # noqa: C901
        self,
        call: Callable[..., Any],
        scope: Scope | None,
        interface: Any = NOT_SET,
        override: bool = False,
        /,
        **defaults: Any,
    ) -> Provider:
        """Register a provider with the specified scope."""
        name = type_repr(call)
        kind = ProviderKind.from_call(call)
        detected_scope = scope

        # Validate scope if it provided
        if scope:
            self._validate_provider_scope(scope, name, kind)

        # Get the signature
        globalns = getattr(call, "__globals__", {})
        module = getattr(call, "__module__", None)
        signature = inspect.signature(call, globals=globalns)

        # Detect the interface
        if interface is NOT_SET:
            if kind == ProviderKind.CLASS:
                interface = call
            else:
                interface = signature.return_annotation
                if interface is inspect.Signature.empty:
                    interface = None

        if isinstance(interface, str):
            interface = get_typed_annotation(interface, globalns, module)

        # If the callable is an iterator, return the actual type
        if is_iterator_type(interface) or is_iterator_type(get_origin(interface)):
            if args := get_args(interface):
                interface = args[0]
                # If the callable is a generator, return the resource type
                if is_none_type(interface):
                    interface = type(f"Event_{uuid.uuid4().hex}", (Event,), {})
            else:
                raise TypeError(
                    f"Cannot use `{name}` resource type annotation "
                    "without actual type argument."
                )

        # None interface is not allowed
        if is_none_type(interface):
            raise TypeError(f"Missing `{name}` provider return annotation.")

        # Check for existing provider
        if interface in self._providers and not override:
            raise LookupError(
                f"The provider interface `{type_repr(interface)}` already registered."
            )

        unresolved_parameter = None
        parameters: list[inspect.Parameter] = []
        scopes: dict[Scope, Provider] = {}

        for parameter in signature.parameters.values():
            if parameter.annotation is inspect.Parameter.empty:
                raise TypeError(
                    f"Missing provider `{name}` "
                    f"dependency `{parameter.name}` annotation."
                )
            if parameter.kind == inspect.Parameter.POSITIONAL_ONLY:
                raise TypeError(
                    "Positional-only parameters "
                    f"are not allowed in the provider `{name}`."
                )

            parameter = parameter.replace(
                annotation=get_typed_annotation(parameter.annotation, globalns, module)
            )

            try:
                sub_provider = self._get_or_register_provider(
                    parameter.annotation, scope
                )
            except LookupError:
                if self._parameter_has_default(parameter, **defaults):
                    continue
                unresolved_parameter = parameter
                continue

            # Store first provider for each scope
            if sub_provider.scope not in scopes:
                scopes[sub_provider.scope] = sub_provider

            parameters.append(parameter)

        # Set detected scope
        if detected_scope is None:
            if "transient" in scopes:
                detected_scope = "transient"
            elif "request" in scopes:
                detected_scope = "request"
            elif "singleton" in scopes:
                detected_scope = "singleton"
            else:
                detected_scope = self.default_scope

        # Validate the provider scope after detection
        if scope is None:
            self._validate_provider_scope(detected_scope, name, kind)

        # Check for unresolved parameters
        if unresolved_parameter:
            if detected_scope not in ("singleton", "transient"):
                self._unresolved_interfaces.add(interface)
            else:
                raise LookupError(
                    f"The provider `{name}` depends on `{unresolved_parameter.name}` "
                    f"of type `{type_repr(unresolved_parameter.annotation)}`, "
                    "which has not been registered or set. To resolve this, ensure "
                    f"that `{unresolved_parameter.name}` is registered before "
                    f"attempting to use it."
                ) from None

        # Check scope compatibility
        for sub_provider in scopes.values():
            if sub_provider.scope not in ALLOWED_SCOPES.get(detected_scope, []):
                raise ValueError(
                    f"The provider `{name}` with a `{detected_scope}` scope cannot "
                    f"depend on `{sub_provider}` with a `{sub_provider.scope}` scope. "
                    "Please ensure all providers are registered with matching scopes."
                )

        provider = Provider(
            call=call,
            scope=detected_scope,
            interface=interface,
            name=name,
            kind=kind,
            parameters=parameters,
        )

        self._set_provider(provider)
        return provider

    @staticmethod
    def _validate_provider_scope(scope: Scope, name: str, kind: ProviderKind) -> None:
        """Validate the provider scope."""
        if scope not in (allowed_scopes := get_args(Scope)):
            raise ValueError(
                f"The provider `{name}` scope is invalid. Only the following "
                f"scopes are supported: {', '.join(allowed_scopes)}. "
                "Please use one of the supported scopes when registering a provider."
            )
        if scope == "transient" and ProviderKind.is_resource(kind):
            raise TypeError(
                f"The resource provider `{name}` is attempting to register "
                "with a transient scope, which is not allowed."
            )

    def _get_provider(self, interface: Any) -> Provider:
        """Get provider by interface."""
        try:
            return self._providers[interface]
        except KeyError as exc:
            raise LookupError(
                f"The provider interface for `{type_repr(interface)}` has "
                "not been registered. Please ensure that the provider interface is "
                "properly registered before attempting to use it."
            ) from exc

    def _get_or_register_provider(
        self, interface: Any, parent_scope: Scope | None, /, **defaults: Any
    ) -> Provider:
        """Get or register a provider by interface."""
        try:
            return self._providers[interface]
        except KeyError:
            if inspect.isclass(interface) and not is_builtin_type(interface):
                # Try to get defined scope
                if is_provided(interface):
                    scope = interface.__provided__["scope"]
                else:
                    scope = parent_scope
                return self._register_provider(interface, scope, interface, **defaults)
            raise LookupError(
                f"The provider interface `{type_repr(interface)}` is either not "
                "registered, not provided, or not set in the scoped context. "
                "Please ensure that the provider interface is properly registered and "
                "that the class is decorated with a scope before attempting to use it."
            ) from None

    def _set_provider(self, provider: Provider) -> None:
        """Set a provider by interface."""
        self._providers[provider.interface] = provider
        if provider.is_resource:
            self._resources[provider.scope].append(provider.interface)

    def _delete_provider(self, provider: Provider) -> None:
        """Delete a provider."""
        if provider.interface in self._providers:
            del self._providers[provider.interface]
        if provider.is_resource:
            self._resources[provider.scope].remove(provider.interface)

    @staticmethod
    def _parameter_has_default(
        parameter: inspect.Parameter, /, **defaults: Any
    ) -> bool:
        has_default_in_kwargs = parameter.name in defaults if defaults else False
        has_default = parameter.default is not inspect.Parameter.empty
        return has_default_in_kwargs or has_default

    ############################
    # Instance Methods
    ############################

    @overload
    def resolve(self, interface: type[T]) -> T: ...

    @overload
    def resolve(self, interface: T) -> T: ...  # type: ignore

    def resolve(self, interface: type[T]) -> T:
        """Resolve an instance by interface."""
        return self._resolve_or_create(interface, False)

    @overload
    async def aresolve(self, interface: type[T]) -> T: ...

    @overload
    async def aresolve(self, interface: T) -> T: ...

    async def aresolve(self, interface: type[T]) -> T:
        """Resolve an instance by interface asynchronously."""
        return await self._aresolve_or_create(interface, False)

    def create(self, interface: type[T], /, **defaults: Any) -> T:
        """Create an instance by interface."""
        return self._resolve_or_create(interface, True, **defaults)

    async def acreate(self, interface: type[T], /, **defaults: Any) -> T:
        """Create an instance by interface asynchronously."""
        return await self._aresolve_or_create(interface, True, **defaults)

    def is_resolved(self, interface: Any) -> bool:
        """Check if an instance by interface exists."""
        try:
            provider = self._get_provider(interface)
        except LookupError:
            return False
        if provider.scope == "transient":
            return False
        context = self._get_instance_context(provider.scope)
        return interface in context

    def release(self, interface: Any) -> None:
        """Release an instance by interface."""
        provider = self._get_provider(interface)
        if provider.scope == "transient":
            return None
        context = self._get_instance_context(provider.scope)
        del context[interface]

    def reset(self) -> None:
        """Reset resolved instances."""
        for interface, provider in self._providers.items():
            if provider.scope == "transient":
                continue
            try:
                context = self._get_instance_context(provider.scope)
            except LookupError:
                continue
            del context[interface]

    def _resolve_or_create(
        self, interface: Any, create: bool, /, **defaults: Any
    ) -> Any:
        """Internal method to handle instance resolution and creation."""
        provider = self._get_or_register_provider(interface, None, **defaults)
        if provider.scope == "transient":
            return self._create_instance(provider, None, **defaults)
        context = self._get_instance_context(provider.scope)
        with context.lock():
            return (
                self._get_or_create_instance(provider, context)
                if not create
                else self._create_instance(provider, context, **defaults)
            )

    async def _aresolve_or_create(
        self, interface: Any, create: bool, /, **defaults: Any
    ) -> Any:
        """Internal method to handle instance resolution and creation asynchronously."""
        provider = self._get_or_register_provider(interface, None, **defaults)
        if provider.scope == "transient":
            return await self._acreate_instance(provider, None, **defaults)
        context = self._get_instance_context(provider.scope)
        async with context.alock():
            return (
                await self._aget_or_create_instance(provider, context)
                if not create
                else await self._acreate_instance(provider, context, **defaults)
            )

    def _get_or_create_instance(
        self, provider: Provider, context: InstanceContext
    ) -> Any:
        """Get an instance of a dependency from the scoped context."""
        instance = context.get(provider.interface)
        if instance is None:
            instance = self._create_instance(provider, context)
            context.set(provider.interface, instance)
            return instance
        return instance

    async def _aget_or_create_instance(
        self, provider: Provider, context: InstanceContext
    ) -> Any:
        """Get an async instance of a dependency from the scoped context."""
        instance = context.get(provider.interface)
        if instance is None:
            instance = await self._acreate_instance(provider, context)
            context.set(provider.interface, instance)
            return instance
        return instance

    def _create_instance(
        self, provider: Provider, context: InstanceContext | None, /, **defaults: Any
    ) -> Any:
        """Create an instance using the provider."""
        if provider.is_async:
            raise TypeError(
                f"The instance for the provider `{provider}` cannot be created in "
                "synchronous mode."
            )

        provider_kwargs = self._get_provided_kwargs(provider, context, **defaults)

        if provider.is_generator:
            if context is None:
                raise ValueError("The context is required for generator providers.")
            cm = contextlib.contextmanager(provider.call)(**provider_kwargs)
            return context.enter(cm)

        instance = provider.call(**provider_kwargs)
        if context is not None and provider.is_class and is_context_manager(instance):
            context.enter(instance)
        return instance

    async def _acreate_instance(
        self, provider: Provider, context: InstanceContext | None, /, **defaults: Any
    ) -> Any:
        """Create an instance asynchronously using the provider."""
        provider_kwargs = await self._aget_provided_kwargs(
            provider, context, **defaults
        )

        if provider.is_coroutine:
            return await provider.call(**provider_kwargs)

        if provider.is_async_generator:
            if context is None:
                raise ValueError(
                    "The async stack is required for async generator providers."
                )
            cm = contextlib.asynccontextmanager(provider.call)(**provider_kwargs)
            return await context.aenter(cm)

        if provider.is_generator:

            def _create() -> Any:
                if context is None:
                    raise ValueError("The stack is required for generator providers.")
                cm = contextlib.contextmanager(provider.call)(**provider_kwargs)
                return context.enter(cm)

            return await run_sync(_create)

        instance = await run_sync(provider.call, **provider_kwargs)
        if (
            context is not None
            and provider.is_class
            and is_async_context_manager(instance)
        ):
            await context.aenter(instance)
        return instance

    def _get_provided_kwargs(
        self, provider: Provider, context: InstanceContext | None, /, **defaults: Any
    ) -> dict[str, Any]:
        """Retrieve the arguments for a provider."""
        provided_kwargs = {}
        for parameter in provider.parameters:
            provided_kwargs[parameter.name] = self._get_provider_instance(
                provider, parameter, context, **defaults
            )
        return {**defaults, **provided_kwargs}

    def _get_provider_instance(
        self,
        provider: Provider,
        parameter: inspect.Parameter,
        context: InstanceContext | None,
        /,
        **defaults: Any,
    ) -> Any:
        """Retrieve an instance of a dependency from the scoped context."""

        # Try to get instance from defaults
        if parameter.name in defaults:
            return defaults[parameter.name]

        # Try to get instance from context
        elif context and parameter.annotation in context:
            instance = context[parameter.annotation]

        # Resolve new instance
        else:
            try:
                instance = self._resolve_parameter(provider, parameter)
            except LookupError:
                if parameter.default is inspect.Parameter.empty:
                    raise
                return parameter.default
        return instance

    async def _aget_provided_kwargs(
        self, provider: Provider, context: InstanceContext | None, /, **defaults: Any
    ) -> dict[str, Any]:
        """Asynchronously retrieve the arguments for a provider."""
        provided_kwargs = {}
        for parameter in provider.parameters:
            provided_kwargs[parameter.name] = await self._aget_provider_instance(
                provider, parameter, context, **defaults
            )
        return {**defaults, **provided_kwargs}

    async def _aget_provider_instance(
        self,
        provider: Provider,
        parameter: inspect.Parameter,
        context: InstanceContext | None,
        /,
        **defaults: Any,
    ) -> Any:
        """Asynchronously retrieve an instance of a dependency from the context."""

        # Try to get instance from defaults
        if parameter.name in defaults:
            return defaults[parameter.name]

        # Try to get instance from context
        elif context and parameter.annotation in context:
            instance = context[parameter.annotation]

        # Resolve new instance
        else:
            try:
                instance = await self._aresolve_parameter(provider, parameter)
            except LookupError:
                if parameter.default is inspect.Parameter.empty:
                    raise
                return parameter.default
        return instance

    def _resolve_parameter(
        self, provider: Provider, parameter: inspect.Parameter
    ) -> Any:
        self._validate_resolvable_parameter(provider, parameter)
        return self._resolve_or_create(parameter.annotation, False)

    async def _aresolve_parameter(
        self, provider: Provider, parameter: inspect.Parameter
    ) -> Any:
        self._validate_resolvable_parameter(provider, parameter)
        return await self._aresolve_or_create(parameter.annotation, False)

    def _validate_resolvable_parameter(
        self, provider: Provider, parameter: inspect.Parameter
    ) -> None:
        """Ensure that the specified interface is resolved."""
        if parameter.annotation in self._unresolved_interfaces:
            raise LookupError(
                f"You are attempting to get the parameter `{parameter.name}` with the "
                f"annotation `{type_repr(parameter.annotation)}` as a "
                f"dependency into `{type_repr(provider.call)}` which is "
                "not registered or set in the scoped context."
            )

    ############################
    # Injector Methods
    ############################

    @overload
    def inject(self, func: Callable[P, T]) -> Callable[P, T]: ...

    @overload
    def inject(self) -> Callable[[Callable[P, T]], Callable[P, T]]: ...

    def inject(
        self, func: Callable[P, T] | None = None
    ) -> Callable[[Callable[P, T]], Callable[P, T]] | Callable[P, T]:
        """Decorator to inject dependencies into a callable."""

        def decorator(call: Callable[P, T]) -> Callable[P, T]:
            return self._inject(call)

        if func is None:
            return decorator
        return decorator(func)

    def run(self, func: Callable[P, T], /, *args: P.args, **kwargs: P.kwargs) -> T:
        """Run the given function with injected dependencies."""
        return self._inject(func)(*args, **kwargs)

    def _inject(self, call: Callable[P, T]) -> Callable[P, T]:
        """Inject dependencies into a callable."""
        if call in self._inject_cache:
            return cast(Callable[P, T], self._inject_cache[call])

        injected_params = self._get_injected_params(call)
        if not injected_params:
            self._inject_cache[call] = call
            return call

        if inspect.iscoroutinefunction(call):

            @functools.wraps(call)
            async def awrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                for name, annotation in injected_params.items():
                    kwargs[name] = await self.aresolve(annotation)
                return cast(T, await call(*args, **kwargs))

            self._inject_cache[call] = awrapper

            return awrapper  # type: ignore

        @functools.wraps(call)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            for name, annotation in injected_params.items():
                kwargs[name] = self.resolve(annotation)
            return call(*args, **kwargs)

        self._inject_cache[call] = wrapper

        return wrapper

    def _get_injected_params(self, call: Callable[..., Any]) -> dict[str, Any]:
        """Get the injected parameters of a callable object."""
        injected_params: dict[str, Any] = {}
        for parameter in get_typed_parameters(call):
            if not is_marker(parameter.default):
                continue
            self._validate_injected_parameter(call, parameter)
            injected_params[parameter.name] = parameter.annotation
        return injected_params

    def _validate_injected_parameter(
        self, call: Callable[..., Any], parameter: inspect.Parameter
    ) -> None:
        """Validate an injected parameter."""
        # TODO: temporary disable until strict is enforced
        return None

        if parameter.annotation is inspect.Parameter.empty:
            raise TypeError(
                f"Missing `{type_repr(call)}` parameter `{parameter.name}` annotation."
            )

        if not self.has_provider_for(parameter.annotation):
            raise LookupError(
                f"`{type_repr(call)}` has an unknown dependency parameter "
                f"`{parameter.name}` with an annotation of "
                f"`{type_repr(parameter.annotation)}`."
            )

    ############################
    # Module Methods
    ############################

    def register_module(self, module: ModuleDef) -> None:
        """Register a module as a callable, module type, or module instance."""
        self._modules.register(module)

    ############################
    # Scanner Methods
    ############################

    def scan(
        self, /, packages: PackageOrIterable, *, tags: Iterable[str] | None = None
    ) -> None:
        self._scanner.scan(packages=packages, tags=tags)



from dataclasses import dataclass

from anydi import auto, injectable

from tests.fixtures import Service


@injectable
def b_handler(service: Service = auto) -> None:
    pass


class BClassHandler:
    def __init__(self, service: Service = auto) -> None:
        self.service = service


@dataclass
class BDataclassHandler:
    service: Service = auto



from typing import Any

import pytest
from faststream.redis import RedisBroker

from anydi import Container
from anydi.ext.faststream import Inject, install


def test_inject_param_missing_interface() -> None:
    param = Inject()

    with pytest.raises(TypeError, match="Interface is not set."):
        _ = param.interface


@pytest.mark.skip(reason="disable until strict is enforced")
def test_install_without_annotation() -> None:
    container = Container()

    @container.provider(scope="singleton")
    def message() -> str:
        return "Hello"

    broker = RedisBroker()

    @broker.subscriber("hello")
    def say_hello(message=Inject()) -> Any:  # type: ignore[no-untyped-def]
        return message

    with pytest.raises(
        TypeError, match="Missing `(.*?).say_hello` parameter `message` annotation."
    ):
        install(broker, container)


@pytest.mark.skip(reason="disable until strict is enforced")
def test_install_unknown_annotation() -> None:
    container = Container()

    broker = RedisBroker()

    @broker.subscriber("hello")
    def say_hello(message: str = Inject()) -> Any:
        return message

    with pytest.raises(
        LookupError,
        match=(
            "`(.*?).say_hello` has an unknown dependency parameter `message` "
            "with an annotation of `str`."
        ),
    ):
        install(broker, container)
